COMPONENT_KEY = "digitizer"


class ModelTypes:
    IMAGE_PROCESSOR = "image_processor"


class StatusKeys:
    DOWNLOAD_MODELS = "digitizer_download_models"
    CLEAN_UP = "digitizer_clean_up"
    ELASTICSEARCH_UPLOAD = "digitizer_elasticsearch_upload"
    UPLOAD = "s3_upload"
    DOWNLOAD = "digitizer_s3_download"
    OCR = "digitizer_ocr"


class Queue:
    IO = "io"
    DOWNLOAD = "download"
    FINISH = "finish"
    OCR = "ocr"


class Tasks:
    MODEL_UPDATE = "component_model_update"
